import functools
import operator
import os
import re
try:
    import simplejson as json
except ImportError:
    import json

from flask import Blueprint, render_template, abort, request, url_for, redirect, flash, Response
from flask_peewee.forms import BooleanSelectField, ForeignKeyField, CustomModelConverter
from flask_peewee.filters import QueryFilter, lookups_for_field, Lookup, FIELD_TYPES
from flask_peewee.serializer import Serializer
from flask_peewee.utils import get_next, PaginatedQuery, path_to_models, slugify
from peewee import BooleanField, DateTimeField, ForeignKeyField, DateField
from werkzeug import Headers
from wtforms import fields, widgets
from wtfpeewee.fields import ModelSelectField, ModelSelectMultipleField
from wtfpeewee.orm import model_form


current_dir = os.path.dirname(__file__)


class FieldValueMap(object):
    def __init__(self, model_admin):
        self.model_admin = model_admin

    def get_conversions(self):
        return {
            (BooleanField, 'eq'): self.handle_boolean,
            (DateTimeField, 'today'): self.handle_dummy,
            (DateTimeField, 'yesterday'): self.handle_dummy,
            (DateTimeField, 'this_week'): self.handle_dummy,
            (DateField, 'today'): self.handle_dummy,
            (DateField, 'yesterday'): self.handle_dummy,
            (DateField, 'this_week'): self.handle_dummy,
            (ForeignKeyField, 'eq'): self.handle_foreign_key,
            (ForeignKeyField, 'in'): self.handle_foreign_key_multi,
        }

    def convert(self, field, lookup, lookup_name, prefix):
        conversions = self.get_conversions()
        key = (type(field), lookup)
        lookup_obj = conversions.get(key, self.handle_default)(field, lookup, prefix)

        # wat
        lookup_obj.lookup_name = lookup_name
        lookup_obj.css_class = 'span2 input-small lookup-input'
        return lookup_obj

    def handle_default(self, field, lookup, prefix):
        return Lookup(field, lookup, 'text', prefix=prefix)

    def handle_dummy(self, field, lookup, prefix):
        return Lookup(field, lookup, 'hidden', prefix=prefix)

    def handle_boolean(self, field, lookup, prefix):
        return Lookup(field, lookup, 'select', [('1', 'True'), ('', 'False')], prefix=prefix)

    def fk_to_list(self, field):
        return [(obj.get_pk(), unicode(obj)) for obj in field.to.select()]

    def handle_foreign_key(self, field, lookup, prefix):
        _fkl = self.model_admin.foreign_key_lookups or {}
        if field.name not in _fkl:
            return Lookup(field, lookup, 'select', self.fk_to_list(field), prefix=prefix)
        return Lookup(field, lookup, 'foreign_key', field.to, prefix=prefix)

    def handle_foreign_key_multi(self, field, lookup, prefix):
        _fkl = self.model_admin.foreign_key_lookups or {}
        if field.name not in _fkl:
            return Lookup(field, lookup, 'select_multiple', self.fk_to_list(field), prefix=prefix)
        return Lookup(field, lookup, 'foreign_key_multiple', field.to, prefix=prefix)


class ModelAdmin(object):
    """
    ModelAdmin provides create/edit/delete functionality for a peewee Model.
    """
    paginate_by = 20
    filter_paginate_by = 15

    # columns to display in the list index - can be field names or callables on
    # a model instance, though in the latter case they will not be sortable
    columns = None

    # filter parameters -- what is available for filtering via the request
    exclude_filters = None
    ignore_filters = ('ordering', 'page',)

    # form parameters, lists of fields
    exclude = None
    fields = None

    # foreign_key_field --> related field to search on, e.g. {'user': 'username'}
    foreign_key_lookups = None

    # delete behavior
    delete_collect_objects = True
    delete_recursive = True

    # templates, to override see get_template_overrides()
    base_templates = {
        'index': 'admin/models/index.html',
        'add': 'admin/models/add.html',
        'edit': 'admin/models/edit.html',
        'delete': 'admin/models/delete.html',
        'export': 'admin/models/export.html',
    }

    def __init__(self, admin, model):
        self.admin = admin
        self.model = model
        self.pk_name = self.model._meta.pk_name

        self.templates = dict(self.base_templates)
        self.templates.update(self.get_template_overrides())

    def get_template_overrides(self):
        return {}

    def get_url_name(self, name):
        return '%s.%s_%s' % (
            self.admin.blueprint.name,
            self.get_admin_name(),
            name,
        )

    def get_form(self):
        return model_form(self.model, only=self.fields, exclude=self.exclude, converter=CustomModelConverter(self))

    def get_add_form(self):
        return self.get_form()

    def get_edit_form(self):
        return self.get_form()

    def get_query(self):
        return self.model.select()

    def get_object(self, pk):
        return self.get_query().get(**{self.pk_name: pk})

    def get_query_filter(self, query):
        return QueryFilter(query, self.ignore_filters)

    def get_urls(self):
        return (
            ('/', self.index),
            ('/add/', self.add),
            ('/delete/', self.delete),
            ('/export/', self.export),
            ('/<pk>/', self.edit),
            ('/_ajax/', self.ajax_list),
        )

    def get_columns(self):
        return self.model._meta.get_field_names()

    def column_is_sortable(self, col):
        return col in self.model._meta.fields

    def get_display_name(self):
        return self.model.__name__

    def get_admin_name(self):
        return slugify(self.model.__name__)

    def get_lookups(self, prefix='', seen=None):
        field_value_map = FieldValueMap(self)

        seen = seen or set()
        lookups = {}
        active_lookups = []

        for field in self.model._meta.get_fields():
            if self.exclude_filters and field.name in self.exclude_filters:
                continue

            for lookup, lookup_name in lookups_for_field(field):
                key = (prefix, field)
                lookups.setdefault(key, [])
                lookup_obj = field_value_map.convert(field, lookup, lookup_name, prefix)
                lookups[key].append(lookup_obj)

                if lookup_obj.name in request.args:
                    active_lookups.append(lookup_obj)

            if isinstance(field, ForeignKeyField):
                rel_prefix = '%s%s__' % (prefix, field.name)
                rel_model = field.to
                if rel_model in self.admin and (rel_model, field.name) not in seen:
                    seen.add((rel_model, field.name))
                    rel_lookups, rel_active = self.admin[rel_model].get_lookups(rel_prefix, seen)
                    lookups.update(rel_lookups)
                    active_lookups.extend(rel_active)

        return lookups, active_lookups

    def save_model(self, instance, form, adding=False):
        form.populate_obj(instance)
        instance.save()
        return instance

    def apply_ordering(self, query, ordering):
        if ordering:
            desc, column = ordering.startswith('-'), ordering.lstrip('-')
            if self.column_is_sortable(column):
                query = query.order_by((column, desc and 'desc' or 'asc'))
        return query

    def index(self):
        query = self.get_query()

        ordering = request.args.get('ordering') or ''
        query = self.apply_ordering(query, ordering)

        # create a QueryFilter object with our current query
        query_filter = self.get_query_filter(query)

        # process the filters from the request
        filtered_query = query_filter.get_filtered_query()

        # create a paginated query out of our filtered results
        pq = PaginatedQuery(filtered_query, self.paginate_by)

        if request.method == 'POST':
            id_list = request.form.getlist('id')
            if request.form['action'] == 'delete':
                return redirect(url_for(self.get_url_name('delete'), id=id_list))
            else:
                return redirect(url_for(self.get_url_name('export'), id__in=id_list))

        lookups, active_lookups = self.get_lookups()

        return render_template(self.templates['index'],
            model_admin=self,
            query=pq,
            ordering=ordering,
            query_filter=query_filter,
            lookups=lookups,
            active_lookups=active_lookups,
        )

    def dispatch_save_redirect(self, instance):
        if 'save' in request.form:
            return redirect(url_for(self.get_url_name('index')))
        elif 'save_add' in request.form:
            return redirect(url_for(self.get_url_name('add')))
        else:
            return redirect(
                url_for(self.get_url_name('edit'), pk=instance.get_pk())
            )

    def add(self):
        Form = self.get_add_form()

        if request.method == 'POST':
            form = Form(request.form)
            if form.validate():
                instance = self.save_model(self.model(), form, True)
                flash('New %s saved successfully' % self.get_display_name(), 'success')
                return self.dispatch_save_redirect(instance)
        else:
            form = Form()

        return render_template(self.templates['add'], model_admin=self, form=form)

    def edit(self, pk):
        try:
            instance = self.get_object(pk)
        except self.model.DoesNotExist:
            abort(404)

        Form = self.get_edit_form()

        if request.method == 'POST':
            form = Form(request.form, obj=instance)
            if form.validate():
                self.save_model(instance, form, False)
                flash('Changes to %s saved successfully' % self.get_display_name(), 'success')
                return self.dispatch_save_redirect(instance)
        else:
            form = Form(obj=instance)

        return render_template(self.templates['edit'], model_admin=self, instance=instance, form=form)

    def collect_objects(self, obj):
        select_queries, nullable_queries = obj.collect_queries()
        objects = []

        for query, fk_field, depth in select_queries:
            model = query.model
            query.query = ['*']
            collected = [obj for obj in query.execute().iterator()]
            if collected:
                objects.append((depth, model, fk_field, collected))

        return sorted(objects, key=lambda i: (i[0], i[1].__name__))

    def delete(self):
        if request.method == 'GET':
            id_list = request.args.getlist('id')
        else:
            id_list = request.form.getlist('id')

        query = self.model.select().where(**{
            '%s__in' % self.model._meta.pk_name: id_list
        })

        if request.method == 'GET':
            collected = {}
            if self.delete_collect_objects:
                for obj in query:
                    collected[obj.get_pk()] = self.collect_objects(obj)

        elif request.method == 'POST':
            count = query.count()
            for obj in query:
                obj.delete_instance(recursive=self.delete_recursive)

            flash('Successfully deleted %s %ss' % (count, self.get_display_name()), 'success')
            return redirect(url_for(self.get_url_name('index')))

        return render_template(self.templates['delete'], **dict(
            model_admin=self,
            query=query,
            collected=collected,
        ))

    def collect_related_fields(self, model, accum, path):
        path_str = '__'.join(path)
        for field in model._meta.get_fields():
            if isinstance(field, ForeignKeyField):
                self.collect_related_fields(field.to, accum, path + [field.name])
            elif model != self.model:
                accum.setdefault((model, path_str), [])
                accum[(model, path_str)].append(field)

        return accum

    def export(self):
        query = self.get_query()

        ordering = request.args.get('ordering') or ''
        query = self.apply_ordering(query, ordering)

        # create a QueryFilter object with our current query
        query_filter = self.get_query_filter(query)

        # process the filters from the request
        filtered_query = query_filter.get_filtered_query()

        related = self.collect_related_fields(self.model, {}, [])

        if request.method == 'POST':
            raw_fields = request.form.getlist('fields')
            export = Export(filtered_query, related, raw_fields)
            return export.json_response('export-%s.json' % self.get_admin_name())

        lookups, active_lookups = self.get_lookups()

        return render_template(self.templates['export'],
            model_admin=self,
            model=filtered_query.model,
            query=filtered_query,
            query_filter=query_filter,
            related_fields=related,
            lookups=lookups,
            active_lookups=active_lookups,
        )

    def ajax_list(self):
        field = request.args.get('field')
        prev_page = 0
        next_page = 0

        try:
            models = path_to_models(self.model, field)
        except AttributeError:
            data = []
        else:
            rel_model = models.pop()
            rel_field = self.foreign_key_lookups[field]

            query = rel_model.select().where(**{
                '%s__icontains'% rel_field: request.args.get('query', ''),
            }).order_by(rel_field)

            pq = PaginatedQuery(query, self.filter_paginate_by)
            current_page = pq.get_page()
            if current_page > 1:
                prev_page = current_page - 1
            if current_page < pq.get_pages():
                next_page = current_page + 1

            data = [
                {'id': obj.get_pk(), 'repr': unicode(obj)} \
                    for obj in pq.get_list()
            ]

        json_data = json.dumps({'prev_page': prev_page, 'next_page': next_page, 'object_list': data})
        return Response(json_data, mimetype='application/json')


class AdminPanel(object):
    template_name = 'admin/panels/default.html'

    def __init__(self, admin, title):
        self.admin = admin
        self.title = title
        self.slug = slugify(self.title)

    def dashboard_url(self):
        return url_for('%s.index' % (self.admin.blueprint.name))

    def get_urls(self):
        return ()

    def get_url_name(self, name):
        return '%s.panel_%s_%s' % (
            self.admin.blueprint.name,
            self.slug,
            name,
        )

    def get_template_name(self):
        return self.template_name

    def get_context(self):
        return {}

    def render(self):
        return render_template(self.get_template_name(), panel=self, **self.get_context())


class AdminTemplateHelper(object):
    def __init__(self, admin):
        self.admin = admin
        self.app = self.admin.app

    def get_model_field(self, model, field):
        attr = getattr(model, field)
        if callable(attr):
            return attr()
        return attr

    def get_form_field(self, form, field_name):
        return getattr(form, field_name)

    def fix_underscores(self, s):
        return s.replace('_', ' ').title()

    def update_querystring(self, querystring, key, val):
        if not querystring:
            return '%s=%s' % (key, val)
        else:
            querystring = re.sub('%s(?:[^&]+)?&?' % key, '', querystring).rstrip('&')
            return ('%s&%s=%s' % (querystring, key, val)).lstrip('&')

    def get_verbose_name(self, model, column_name):
        try:
            field = model._meta.fields[column_name]
        except KeyError:
            return self.fix_underscores(column_name)
        else:
            return field.verbose_name

    def get_model_admins(self):
        return {'model_admins': self.admin.get_model_admins()}

    def get_admin_url(self, obj):
        model_admin = self.admin.get_admin_for(type(obj))
        if model_admin:
            return url_for(model_admin.get_url_name('edit'), pk=obj.get_pk())

    def get_model_name(self, model_class):
        model_admin = self.admin.get_admin_for(model_class)
        if model_admin:
            return model_admin.get_display_name()
        return model_class.__name__

    def prepare_environment(self):
        self.app.template_context_processors[None].append(self.get_model_admins)

        self.app.jinja_env.globals['get_model_field'] = self.get_model_field
        self.app.jinja_env.globals['get_form_field'] = self.get_form_field
        self.app.jinja_env.globals['get_verbose_name'] = self.get_verbose_name
        self.app.jinja_env.filters['fix_underscores'] = self.fix_underscores
        self.app.jinja_env.globals['update_querystring'] = self.update_querystring
        self.app.jinja_env.globals['get_admin_url'] = self.get_admin_url
        self.app.jinja_env.globals['get_model_name'] = self.get_model_name


class Admin(object):
    def __init__(self, app, auth, template_helper=AdminTemplateHelper,
                 prefix='/admin', name='admin'):
        self.app = app
        self.auth = auth

        self._admin_models = {}
        self._registry = {}
        self._panels = {}

        self.blueprint = self.get_blueprint(name)
        self.url_prefix = prefix

        self.template_helper = template_helper(self)
        self.template_helper.prepare_environment()

    def auth_required(self, func):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            user = self.auth.get_logged_in_user()

            if not user:
                login_url = url_for('%s.login' % self.auth.blueprint.name, next=get_next())
                return redirect(login_url)

            if not self.check_user_permission(user):
                abort(403)

            return func(*args, **kwargs)
        return inner

    def check_user_permission(self, user):
        return user.admin

    def get_urls(self):
        return (
            ('/', self.auth_required(self.index)),
        )

    def __contains__(self, item):
        return item in self._registry

    def __getitem__(self, item):
        return self._registry[item]

    def register(self, model, admin_class=ModelAdmin):
        model_admin = admin_class(self, model)
        admin_name = model_admin.get_admin_name()

        self._registry[model] = model_admin

    def unregister(self, model):
        del(self._registry[model])

    def register_panel(self, title, panel):
        panel_instance = panel(self, title)
        self._panels[title] = panel_instance

    def unregister_panel(self, title):
        del(self._panels[title])

    def get_admin_for(self, model):
        return self._registry.get(model)

    def get_model_admins(self):
        return sorted(self._registry.values(), key=lambda o: o.get_admin_name())

    def get_panels(self):
        return sorted(self._panels.values(), key=lambda o: o.slug)

    def index(self):
        return render_template('admin/index.html',
            model_admins=self.get_model_admins(),
            panels=self.get_panels(),
        )

    def get_blueprint(self, blueprint_name):
        return Blueprint(
            blueprint_name,
            __name__,
            static_folder=os.path.join(current_dir, 'static'),
            template_folder=os.path.join(current_dir, 'templates'),
        )

    def register_blueprint(self, **kwargs):
        self.app.register_blueprint(
            self.blueprint,
            url_prefix=self.url_prefix,
            **kwargs
        )

    def configure_routes(self):
        for url, callback in self.get_urls():
            self.blueprint.route(url, methods=['GET', 'POST'])(callback)

        for model_admin in self._registry.values():
            admin_name = model_admin.get_admin_name()
            for url, callback in model_admin.get_urls():
                full_url = '/%s%s' % (admin_name, url)
                self.blueprint.add_url_rule(
                    full_url,
                    '%s_%s' % (admin_name, callback.__name__),
                    self.auth_required(callback),
                    methods=['GET', 'POST'],
                )

        for panel in self._panels.values():
            for url, callback in panel.get_urls():
                full_url = '/%s%s' % (panel.slug, url)
                self.blueprint.add_url_rule(
                    full_url,
                    'panel_%s_%s' % (panel.slug, callback.__name__),
                    self.auth_required(callback),
                    methods=['GET', 'POST'],
                )

    def setup(self):
        self.configure_routes()
        self.register_blueprint()


class Export(object):
    def __init__(self, query, related, fields):
        self.query = query
        self.related = related
        self.fields = fields

        self.alias_to_model = dict([(k[1], k[0]) for k in self.related.keys()])

    def prepare_query(self):
        clone = self.query.clone()

        select = {}

        def ensure_join(query, m, p):
            if m not in query._joined_models:
                if '__' not in p:
                    next_model = query.model
                else:
                    next, _ = p.rsplit('__', 1)
                    next_model = self.alias_to_model[next]
                    query = ensure_join(query, next_model, next)

                return query.switch(next_model).join(m)
            else:
                return query

        for lookup in self.fields:
            # lookup may be something like "content" or "user__user_name"
            if '__' in lookup:
                path, column = lookup.rsplit('__', 1)
                model = self.alias_to_model[path]
                clone = ensure_join(clone, model, path)
            else:
                model = self.query.model
                column = lookup

            select.setdefault(model, [])
            select[model].append(column)

        clone.query = select
        return clone

    def json_response(self, filename='export.json'):
        serializer = Serializer()
        prepared_query = self.prepare_query()

        def generate():
            i = prepared_query.count()
            yield '[\n'
            for obj in prepared_query:
                i -= 1
                yield json.dumps(serializer.serialize_object(obj, prepared_query.query))
                if i > 0:
                    yield ',\n'
            yield '\n]'
        headers = Headers()
        headers.add('Content-Type', 'application/javascript')
        headers.add('Content-Disposition', 'attachment; filename=%s' % filename)
        return Response(generate(), mimetype='text/javascript', headers=headers, direct_passthrough=True)



import datetime

from peewee import BooleanField, DateTimeField, ForeignKeyField, TimeField, DateField
from wtforms import fields, form, widgets
from wtforms.fields import FormField, _unset_value
from wtforms.widgets import HTMLString, html_params

from wtfpeewee.fields import ModelSelectField, ModelHiddenField
from wtfpeewee.orm import ModelConverter


class BooleanSelectField(fields.SelectFieldBase):
    widget = widgets.Select()

    def iter_choices(self):
        yield ('1', 'True', self.data)
        yield ('', 'False', not self.data)

    def process_data(self, value):
        try:
            self.data = bool(value)
        except (ValueError, TypeError):
            self.data = None

    def process_formdata(self, valuelist):
        if valuelist:
            try:
                self.data = bool(valuelist[0])
            except ValueError:
                raise ValueError(self.gettext(u'Invalid Choice: could not coerce'))


class CustomTimeField(fields.Field):
    widget = widgets.TextInput()

    def __init__(self, label=None, validators=None, format='%H:%M:%S', **kwargs):
        super(CustomTimeField, self).__init__(label, validators, **kwargs)
        self.format = format

    def _value(self):
        if self.raw_data:
            return u' '.join(self.raw_data)
        else:
            return self.data and self.data.strftime(self.format) or u''

    def process_formdata(self, valuelist):
        if valuelist:
            date_str = u' '.join(valuelist)
            try:
                self.data = datetime.datetime.strptime(date_str, self.format).time()
            except ValueError:
                try:
                    self.data = datetime.datetime.strptime(date_str, '%H:%M').time()
                except ValueError:
                    self.data = None
                    raise ValueError(self.gettext(u'Not a valid time value'))


def inject_class(kwargs, *klasses):
    i_class = list(klasses)
    copy = dict(kwargs)
    if 'class' in copy:
        i_class.append(copy.pop('class'))
    copy['class'] = ' '.join(i_class)
    return copy

def datetime_widget(field, **kwargs):
    kwargs.setdefault('id', field.id)
    html = []
    for subfield in field:
        if isinstance(subfield, fields.DateField):
            kwarg_copy = inject_class(kwargs, 'date-widget', 'datetime-widget')
        elif isinstance(subfield, CustomTimeField):
            kwarg_copy = inject_class(kwargs, 'time-widget', 'datetime-widget')
        html.append(subfield(**kwarg_copy))

    return HTMLString(u''.join(html))

class _DateTimeForm(form.Form):
    date = fields.DateField()
    time = CustomTimeField()

class CustomDateTimeField(FormField):
    widget = staticmethod(datetime_widget)

    def __init__(self, label='', validators=None, **kwargs):
        validators = None
        super(CustomDateTimeField, self).__init__(
            _DateTimeForm, label, validators,
            **kwargs
        )

    def process(self, formdata, data=_unset_value):
        prefix = self.name + self.separator
        kwargs = {}
        if data is _unset_value:
            try:
                data = self.default()
            except TypeError:
                data = self.default
        
        if data and data is not _unset_value:
            kwargs['date'] = data.date()
            kwargs['time'] = data.time()

        self.form = self.form_class(formdata, prefix=prefix, **kwargs)
    
    def populate_obj(self, obj, name):
        setattr(obj, name, self.data)

    @property
    def data(self):
        if self.date.data is not None and self.time.data is not None:
            return datetime.datetime.combine(self.date.data, self.time.data)


class CustomModelConverter(ModelConverter):
    def __init__(self, model_admin, additional=None):
        super(CustomModelConverter, self).__init__(additional)
        self.model_admin = model_admin
        self.converters[BooleanField] = self.handle_boolean
        self.converters[DateTimeField] = self.handle_datetime
        self.converters[TimeField] = self.handle_time
        self.converters[DateField] = self.handle_date

    def handle_boolean(self, model, field, **kwargs):
        return field.name, BooleanSelectField(**kwargs)
    
    def handle_datetime(self, model, field, **kwargs):
        return field.name, CustomDateTimeField(**kwargs)
    
    def handle_time(self, model, field, **kwargs):
        return field.name, CustomTimeField(**kwargs)
    
    def handle_date(self, model, field, **kwargs):
        return field.name, fields.DateField(**inject_class(kwargs, 'date-widget'))

    def handle_foreign_key(self, model, field, **kwargs):
        if field.null:
            kwargs['allow_blank'] = True

        if field.name in (self.model_admin.foreign_key_lookups or ()):
            form_field = ModelHiddenField(model=field.to, **kwargs)
        else:
            form_field = ModelSelectField(model=field.to, **kwargs)
        return field.name, form_field



import math
import random
import re
import sys
from hashlib import sha1

from flask import abort, request, render_template
from peewee import Model, DoesNotExist, SelectQuery, ForeignKeyField


def get_object_or_404(query_or_model, **query):
    try:
        return query_or_model.get(**query)
    except DoesNotExist:
        abort(404)

def object_list(template_name, qr, var_name='object_list', **kwargs):
    pq = PaginatedQuery(qr, kwargs.pop('paginate_by', 20))
    kwargs[var_name] = pq.get_list()
    return render_template(template_name, pagination=pq, page=pq.get_page(), **kwargs)


class PaginatedQuery(object):
    page_var = 'page'

    def __init__(self, query_or_model, paginate_by):
        self.paginate_by = paginate_by

        if isinstance(query_or_model, SelectQuery):
            self.query = query_or_model
            self.model = self.query.model
        else:
            self.model = query_or_model
            self.query = self.model.select()

    def get_page(self):
        curr_page = request.args.get(self.page_var)
        if curr_page and curr_page.isdigit():
            return int(curr_page)
        return 1

    def get_pages(self):
        return math.ceil(float(self.query.count()) / self.paginate_by)

    def get_list(self):
        return self.query.paginate(self.get_page(), self.paginate_by)


def get_next():
    if not request.query_string:
        return request.path
    return '%s?%s' % (request.path, request.query_string)

def slugify(s):
    return re.sub('[^a-z0-9_\-]+', '-', s.lower())

def load_class(s):
    path, klass = s.rsplit('.', 1)
    __import__(path)
    mod = sys.modules[path]
    return getattr(mod, klass)

def get_string_lookups_for_model(model, include_foreign_keys=False, fields=None, exclude=None, accum=None):
    """
    Returns a list of 2-tuples: [
        ('field_a', field_a_obj),
        ('field_b', field_b_obj),
        ('rel__rel_field_a', rel_field_a_obj),
        # ...
    ]

    fields & exclude parameters: {
        Model: [f1, f2],
        RelModel: [rf1, ...],
    }

    this fails though when there are multiple foreign keys to the same model

    perhaps better:
    [f1, f2, {f3: [rf1, rf2]}]
    """
    if isinstance(model, Model):
        model_class = type(model)
    else:
        model_class = model

    lookups = []
    models = [model]

    accum = accum or []

    for field in model._meta.get_fields():
        if isinstance(field, ForeignKeyField):
            rel_model = field.to

            if isinstance(model, Model):
                try:
                    rel_obj = getattr(model, field.name)
                except rel_model.DoesNotExist:
                    rel_obj = None
            else:
                rel_obj = rel_model

            if rel_obj and (not fields or rel_model in fields):
                rel_lookups, rel_models = get_string_lookups_for_model(
                    rel_obj,
                    include_foreign_keys,
                    fields,
                    exclude,
                    accum + [field.name],
                )
                lookups.extend(rel_lookups)
                models.extend(rel_models)

        if include_foreign_keys or not isinstance(field, ForeignKeyField):
            if (not fields or field.name in fields.get(model_class, ())) and \
               (not exclude or (exclude and field.name not in exclude.get(model_class, ()))):
                lookups.append(
                    ('__'.join(accum + [field.name]), getattr(model, field.name))
                )

    return lookups, models

def get_dictionary_lookups_for_model(model, include_foreign_keys=False, fields=None, exclude=None):
    lookups, models = get_string_lookups_for_model(
        model,
        include_foreign_keys,
        fields=fields,
        exclude=exclude,
    )
    return convert_string_lookups_to_dict(lookups), models

def get_models_from_string_lookups(model, lookups):
    """
    Returns a fully-populated model instance from a list of 2-tuples of
    lookup/value: [
        ('field_a', 'value a'),
        ('field_b', 'value_b'),
        ('rel__rel_field_a', 'rel_value_a'),
        # ...
    ]
    """
    field_dict = convert_string_lookups_to_dict(lookups)
    return get_models_from_dictionary(model, field_dict)

def convert_string_lookups_to_dict(lookups):
    field_dict = {}
    split_lookups = [(l.split('__'), v) for l, v in lookups]
    for (lookups, value) in sorted(split_lookups):
        curr = field_dict
        for piece in lookups[:-1]:
            if not isinstance(curr.get(piece, None), dict):
                curr[piece] = {}

            curr = curr[piece]
        curr[lookups[-1]] = value
    return field_dict

def get_models_from_dictionary(model, field_dict):
    if isinstance(model, Model):
        model_instance = model
        check_fks = True
    else:
        model_instance = model()
        check_fks = False
    models = [model_instance]
    for field_name, value in field_dict.items():
        field_obj = model._meta.fields[field_name]
        if isinstance(value, dict):
            rel_obj = field_obj.to
            if check_fks:
                try:
                    rel_obj = getattr(model, field_name)
                except field_obj.to.DoesNotExist:
                    pass
                if rel_obj is None:
                    rel_obj = field_obj.to
            rel_inst, rel_models = get_models_from_dictionary(rel_obj, value)
            models.extend(rel_models)
            setattr(model_instance, field_name, rel_inst)
        else:
            setattr(model_instance, field_name, field_obj.python_value(value))
    return model_instance, models

def path_to_models(model, path):
    accum = []
    if '__' in path:
        next, path = path.split('__')
    else:
        next, path = path, ''
    if next in model._meta.rel_fields:
        field_name = model._meta.rel_fields[next]
        model = model._meta.get_field_by_name(field_name).to
        accum.append(model)
    else:
        raise AttributeError('%s has no related field named "%s"' % (model, next))
    if path:
        accum.extend(path_to_models(model, path))
    return accum

def models_to_path(models):
    accum = []
    last = models[0]
    for model in models[1:]:
        fk_field = last._meta.rel_exists(model)
        if fk_field:
            if fk_field in last._meta.get_fields():
                accum.append(fk_field.name)
            else:
                accum.append(fk_field.related_name)
        else:
            raise AttributeError('%s has no relation to %s' % (last, model))
        last = model
    return '__'.join(accum)


# borrowing these methods, slightly modified, from django.contrib.auth
def get_hexdigest(salt, raw_password):
    return sha1(salt + raw_password).hexdigest()

def make_password(raw_password):
    salt = get_hexdigest(str(random.random()), str(random.random()))[:5]
    hsh = get_hexdigest(salt, raw_password)
    return '%s$%s' % (salt, hsh)

def check_password(raw_password, enc_password):
    salt, hsh = enc_password.split('$', 1)
    return hsh == get_hexdigest(salt, raw_password)



import functools
try:
    import simplejson as json
except ImportError:
    import json

from flask import Blueprint, abort, request, Response, session, redirect, url_for, g
from peewee import *

from flask_peewee.filters import QueryFilter
from flask_peewee.serializer import Serializer, Deserializer
from flask_peewee.utils import PaginatedQuery, slugify, get_object_or_404


class Authentication(object):
    def __init__(self, protected_methods=None):
        if protected_methods is None:
            protected_methods = ['POST', 'PUT', 'DELETE']

        self.protected_methods = protected_methods

    def authorize(self):
        if request.method in self.protected_methods:
            return False

        return True


class APIKeyAuthentication(Authentication):
    """
    Requires a model that has at least two fields, "key" and "secret", which will
    be searched for when authing a request
    """
    def __init__(self, model, protected_methods=None):
        super(APIKeyAuthentication, self).__init__(protected_methods)
        self.model = model

    def get_query(self):
        return self.model.select()

    def get_key(self, k, s):
        try:
            return self.get_query().get(key=k, secret=s)
        except self.model.DoesNotExist:
            pass

    def authorize(self):
        g.api_key = None

        if request.method not in self.protected_methods:
            return True

        if 'key' in request.args and 'secret' in request.args:
            g.api_key = self.get_key(request.args['key'], request.args['secret'])

        return g.api_key


class UserAuthentication(Authentication):
    def __init__(self, auth, protected_methods=None):
        super(UserAuthentication, self).__init__(protected_methods)
        self.auth = auth

    def authorize(self):
        g.user = None

        if request.method not in self.protected_methods:
            return True

        basic_auth = request.authorization
        if not basic_auth:
            return False

        g.user = self.auth.authenticate(basic_auth.username, basic_auth.password)
        return g.user


class AdminAuthentication(UserAuthentication):
    def verify_user(self, user):
        return user.admin

    def authorize(self):
        res = super(AdminAuthentication, self).authorize()

        if res and g.user:
            return self.verify_user(g.user)
        return res


class RestResource(object):
    paginate_by = 20

    # filtering
    exclude_filters = None
    ignore_filters = ('ordering', 'page', 'limit', 'key', 'secret',)

    # serializing: dictionary of model -> field names to restrict output
    fields = None
    exclude = None

    # mapping of field name to resource class
    include_resources = None

    # delete behavior
    delete_recursive = True

    def __init__(self, rest_api, model, authentication, allowed_methods=None):
        self.api = rest_api
        self.model = model
        self.authentication = authentication
        self.allowed_methods = allowed_methods or ['GET', 'POST', 'PUT', 'DELETE']

        self._fields = {self.model: self.fields or self.model._meta.get_field_names()}
        if self.exclude:
            self._exclude = {self.model: self.exclude}
        else:
            self._exclude = {}

        self._resources = {}
        if self.include_resources:
            for field_name, resource in self.include_resources.items():
                field_obj = self.model._meta.fields[field_name]
                resource_obj = resource(self.api, field_obj.to, self.authentication, self.allowed_methods)
                self._resources[field_name] = resource_obj
                self._fields.update(resource_obj._fields)
                self._exclude.update(resource_obj._exclude)

            self._include_foreign_keys = False
        else:
            self._include_foreign_keys = True

    def authorize(self):
        return self.authentication.authorize()

    def get_api_name(self):
        return slugify(self.model.__name__)

    def get_url_name(self, name):
        return '%s.%s_%s' % (
            self.api.blueprint.name,
            self.get_api_name(),
            name,
        )

    def get_query(self):
        return self.model.select()

    def get_query_filter(self, query):
        return QueryFilter(query, self.exclude_filters, self.ignore_filters)

    def get_serializer(self):
        return Serializer()

    def get_deserializer(self):
        return Deserializer()

    def prepare_data(self, obj, data):
        """
        Hook for modifying outgoing data
        """
        return data

    def serialize_object(self, obj):
        s = self.get_serializer()
        return self.prepare_data(
            obj, s.serialize_object(obj, self._fields, self._exclude, self._include_foreign_keys)
        )

    def serialize_query(self, query):
        s = self.get_serializer()
        return [
            self.prepare_data(obj, s.serialize_object(obj, self._fields, self._exclude, self._include_foreign_keys)) \
                for obj in query
        ]

    def deserialize_object(self, data, instance):
        d = self.get_deserializer()
        return d.deserialize_object(instance, data)

    def response_forbidden(self):
        return Response('Forbidden', 403)

    def response_bad_method(self):
        return Response('Unsupported method "%s"' % (request.method), 405)

    def response_bad_request(self):
        return Response('Bad request', 400)

    def response(self, data):
        kwargs = {} if request.is_xhr else {'indent': 2}
        return Response(json.dumps(data, **kwargs), mimetype='application/json')

    def require_method(self, func, methods):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            if request.method not in methods:
                return self.response_bad_method()
            return func(*args, **kwargs)
        return inner

    def get_urls(self):
        return (
            ('/', self.require_method(self.api_list, ['GET', 'POST'])),
            ('/<pk>/', self.require_method(self.api_detail, ['GET', 'POST', 'PUT', 'DELETE'])),
            ('/<pk>/delete/', self.require_method(self.post_delete, ['POST', 'DELETE'])),
        )

    def check_get(self, obj=None):
        return True

    def check_post(self, obj=None):
        return True

    def check_put(self, obj):
        return True

    def check_delete(self, obj):
        return True

    def save_object(self, instance, raw_data):
        instance.save()
        return instance

    def api_list(self):
        if not getattr(self, 'check_%s' % request.method.lower())():
            return self.response_forbidden()

        if request.method == 'GET':
            return self.object_list()
        elif request.method == 'POST':
            return self.create()

    def api_detail(self, pk, method=None):
        obj = get_object_or_404(self.get_query(), **{
            self.model._meta.pk_name: pk
        })

        method = method or request.method

        if not getattr(self, 'check_%s' % method.lower())(obj):
            return self.response_forbidden()

        if method == 'GET':
            return self.object_detail(obj)
        elif method in ('PUT', 'POST'):
            return self.edit(obj)
        elif method == 'DELETE':
            return self.delete(obj)

    def post_delete(self, pk):
        return self.api_detail(pk, 'DELETE')

    def apply_ordering(self, query):
        ordering = request.args.get('ordering') or ''
        if ordering:
            desc, column = ordering.startswith('-'), ordering.lstrip('-')
            if column in self.model._meta.fields:
                query = query.order_by((column, desc and 'desc' or 'asc'))

        return query

    def get_request_metadata(self, paginated_query):
        var = paginated_query.page_var
        request_arguments = request.args.copy()

        current_page = paginated_query.get_page()
        next = previous = ''

        if current_page > 1:
            request_arguments[var] = current_page - 1
            previous = url_for(self.get_url_name('api_list'), **request_arguments)
        if current_page < paginated_query.get_pages():
            request_arguments[var] = current_page + 1
            next = url_for(self.get_url_name('api_list'), **request_arguments)

        return {
            'model': self.get_api_name(),
            'page': current_page,
            'previous': previous,
            'next': next,
        }

    def paginated_object_list(self, filtered_query):
        try:
            paginate_by = int(request.args.get('limit', self.paginate_by))
        except ValueError:
            paginate_by = self.paginate_by
        else:
            if self.paginate_by:
                paginate_by = min(paginate_by, self.paginate_by) # restrict

        pq = PaginatedQuery(filtered_query, paginate_by)
        meta_data = self.get_request_metadata(pq)

        query_dict = self.serialize_query(pq.get_list())

        return self.response({
            'meta': meta_data,
            'objects': query_dict,
        })

    def object_list(self):
        query = self.get_query()
        query = self.apply_ordering(query)

        # create a QueryFilter object with our current query
        query_filter = self.get_query_filter(query)

        # process the filters from the request
        filtered_query = query_filter.get_filtered_query()

        if self.paginate_by or 'limit' in request.args:
            return self.paginated_object_list(filtered_query)

        return self.response(self.serialize_query(filtered_query))

    def object_detail(self, obj):
        return self.response(self.serialize_object(obj))

    def save_related_objects(self, instance, data):
        for k, v in data.items():
            if k in self._resources and isinstance(v, dict):
                rel_resource = self._resources[k]
                rel_obj, rel_models = rel_resource.deserialize_object(v, getattr(instance, k))
                rel_resource.save_related_objects(rel_obj, v)
                setattr(instance, k, rel_resource.save_object(rel_obj, v))

    def create(self):
        data = request.data or request.form.get('data') or ''

        try:
            data = json.loads(data)
        except ValueError:
            return self.response_bad_request()

        instance, models = self.deserialize_object(data, self.model())

        self.save_related_objects(instance, data)
        instance = self.save_object(instance, data)

        return self.response(self.serialize_object(instance))

    def edit(self, obj):
        data = request.data or request.form.get('data') or ''
        try:
            data = json.loads(data)
        except ValueError:
            return self.response_bad_request()

        obj, models = self.deserialize_object(data, obj)

        self.save_related_objects(obj, data)
        obj = self.save_object(obj, data)

        return self.response(self.serialize_object(obj))

    def delete(self, obj):
        res = obj.delete_instance(recursive=self.delete_recursive)
        return self.response({'deleted': res})


class RestrictOwnerResource(RestResource):
    # restrict PUT/DELETE to owner of an object, likewise apply owner to any
    # incoming POSTs
    owner_field = 'user'

    def validate_owner(self, user, obj):
        return user == getattr(obj, self.owner_field)

    def set_owner(self, obj, user):
        setattr(obj, self.owner_field, user)

    def check_put(self, obj):
        return self.validate_owner(g.user, obj)

    def check_delete(self, obj):
        return self.validate_owner(g.user, obj)

    def save_object(self, instance, raw_data):
        self.set_owner(instance, g.user)
        return super(RestrictOwnerResource, self).save_object(instance, raw_data)


class RestAPI(object):
    def __init__(self, app, prefix='/api', default_auth=None, name='api'):
        self.app = app

        self._registry = {}

        self.url_prefix = prefix
        self.blueprint = self.get_blueprint(name)

        self.default_auth = default_auth or Authentication()

    def register(self, model, provider=RestResource, auth=None, allowed_methods=None):
        self._registry[model] = provider(self, model, auth or self.default_auth, allowed_methods)

    def unregister(self, model):
        del(self._registry[model])

    def is_registered(self, model):
        return self._registry.get(model)

    def response_auth_failed(self):
        return Response('Authentication failed', 401, {
            'WWW-Authenticate': 'Basic realm="Login Required"'
        })

    def auth_wrapper(self, func, provider):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            if not provider.authorize():
                return self.response_auth_failed()
            return func(*args, **kwargs)
        return inner

    def get_blueprint(self, blueprint_name):
        return Blueprint(blueprint_name, __name__)

    def get_urls(self):
        return ()

    def configure_routes(self):
        for url, callback in self.get_urls():
            self.blueprint.route(url)(callback)

        for provider in self._registry.values():
            api_name = provider.get_api_name()
            for url, callback in provider.get_urls():
                full_url = '/%s%s' % (api_name, url)
                self.blueprint.add_url_rule(
                    full_url,
                    '%s_%s' % (api_name, callback.__name__),
                    self.auth_wrapper(callback, provider),
                    methods=provider.allowed_methods,
                )

    def register_blueprint(self, **kwargs):
        self.app.register_blueprint(self.blueprint, url_prefix=self.url_prefix, **kwargs)

    def setup(self):
        self.configure_routes()
        self.register_blueprint()



import datetime
import operator

from flask import request
from flask_peewee.utils import models_to_path
from peewee import *


LOOKUP_TYPES = {
    'eq': 'equal to',
    'lt': 'less than',
    'gt': 'greater than',
    'lte': 'less than or equal to',
    'gte': 'greater than or equal to',
    'ne': 'not equal to',
    'istartswith': 'starts with',
    'icontains': 'contains',
    'in': 'is one of',

    # special lookups that require some preprocessing
    'today': 'is today',
    'yesterday': 'is yesterday',
    'this_week': 'this week',
    'lte_days_ago': 'less than ... days ago',
    'gte_days_ago': 'greater than ... days ago',
    'year_eq': 'year is',
    'year_lt': 'year is less than...',
    'year_gt': 'year is greater than...',
}

FIELD_TYPES = {
    'foreign_key': [ForeignKeyField],
    'text': [CharField, TextField],
    'numeric': [PrimaryKeyField, IntegerField, FloatField, DecimalField, DoubleField],
    'boolean': [BooleanField],
    'datetime': [DateTimeField, DateField],
    'time': [TimeField],
}

INV_FIELD_TYPES = dict((v,k) for k in FIELD_TYPES for v in FIELD_TYPES[k])

FIELDS_TO_LOOKUPS = {
    'foreign_key': ['eq', 'in'],
    'text': ['eq', 'icontains', 'istartswith'],
    'numeric': ['eq', 'ne', 'lt', 'lte', 'gt', 'gte', 'in'],
    'boolean': ['eq'],
    'datetime': ['today', 'yesterday', 'this_week', 'lte_days_ago', 'gte_days_ago', 'year_eq', 'year_lt', 'year_gt'],
    'time': [],
}


def is_valid_lookup(model, lookup):
    lookups = lookup.split('__')
    lookup_type = 'eq'

    if len(lookups) > 1 and lookups[-1] in LOOKUP_TYPES:
        lookup_type = lookups.pop()

    curr = model
    for rel_field in lookups[:-1]:
        if rel_field in curr._meta.rel_fields:
            # is this a lookup using the descriptor name of a foreign key field,
            # if so change it to the actual field name, e.g. user -> user_id
            rel_field = curr._meta.rel_fields[rel_field]

        if rel_field in curr._meta.fields:
            field_obj = curr._meta.fields[rel_field]
            if isinstance(field_obj, ForeignKeyField):
                curr = field_obj.to
            else:
                return False
        else:
            if rel_field in curr._meta.reverse_relations:
                curr = curr._meta.reverse_relations[rel_field]
            else:
                return False

    lookup = lookups[-1]
    if lookup in curr._meta.rel_fields:
        lookup = curr._meta.rel_fields[lookup]

    if lookup not in curr._meta.fields:
        return False

    field_obj = curr._meta.fields[lookup]
    if lookup_type not in FIELDS_TO_LOOKUPS[INV_FIELD_TYPES[type(field_obj)]]:
        return False

    return (curr, lookup, lookup_type)

def lookups_for_field(field):
    """
    Get a list of valid lookups for a given field type.  Lookups are expressed
    as a 2-tuple of (<lookup suffix>, <human description>)

    >>> lookups_for_field(User.username)
    [('eq', 'equal to'), ('icontains', 'contains'), ('istartswith', 'starts with')]
    """
    field_class = type(field)
    return [
        (lookup, LOOKUP_TYPES[lookup]) \
            for lookup in FIELDS_TO_LOOKUPS[INV_FIELD_TYPES[field_class]]
    ]

def _rd(n):
    return datetime.date.today() + datetime.timedelta(days=n)

def yr_l(n):
    return datetime.datetime(year=n, month=1, day=1)

def yr_h(n):
    return datetime.datetime.combine(
        datetime.date(year=n, month=12, day=31),
        datetime.time.max,
    )


class FilterPreprocessor(object):
    def process_lookup(self, raw_lookup, values=None):
        """
        Returns a list of dictionaries
        """
        if '__' in raw_lookup:
            field_part, lookup = raw_lookup.rsplit('__', 1)
            if hasattr(self, 'process_%s' % lookup):
                return getattr(self, 'process_%s' % lookup)(field_part, values)

        return [{raw_lookup: v} for v in values]

    def process_in(self, field_part, values):
        return [{'%s__in' % field_part: values}]

    def process_today(self, field_part, values):
        return [{
            '%s__gte' % field_part: _rd(0),
            '%s__lt' % field_part: _rd(1),
        }]

    def process_yesterday(self, field_part, values):
        return [{
            '%s__gte' % field_part: _rd(-1),
            '%s__lt' % field_part: _rd(0),
        }]

    def process_this_week(self, field_part, values):
        return [{
            '%s__gte' % field_part: _rd(-6),
            '%s__lt' % field_part: _rd(1),
        }]

    def process_lte_days_ago(self, field_part, values):
        return [{
            '%s__gte' % field_part: _rd(-1 * int(value)),
        } for value in values]

    def process_gte_days_ago(self, field_part, values):
        return [{
            '%s__lte' % field_part: _rd(-1 * int(value)),
        } for value in values]

    def process_year_eq(self, field_part, values):
        return [{
            '%s__gte' % field_part: yr_l(int(value)),
            '%s__lte' % field_part: yr_h(int(value)),
        } for value in values]

    def process_year_lt(self, field_part, values):
        return [{
            '%s__lt' % field_part: yr_l(int(value)),
        } for value in values]

    def process_year_gt(self, field_part, values):
        return [{
            '%s__gte' % field_part: yr_l(int(value)),
        } for value in values]


class QueryFilter(object):
    def __init__(self, query, exclude=None, ignore_filters=None):
        self.query = query
        self.model = self.query.model

        self.exclude = exclude or ()
        self.ignore_filters = ignore_filters or ()

    def get_preprocessor(self):
        return FilterPreprocessor()

    def process_request(self):
        self.raw_lookups = []

        filters = []
        lookups_by_column = {}

        preprocessor = self.get_preprocessor()

        # preprocessing -- essentially a place to store "raw" filters (used to
        # reconstruct filter widgets on frontend), and a place to munge filters
        # that have special meaning, i.e. "today"
        for key in request.args:
            if key in self.ignore_filters:
                continue

            if not is_valid_lookup(self.model, key):
                continue

            values = request.args.getlist(key)
            if key.endswith('__in'):
                self.raw_lookups.append((key, values))
            else:
                for v in values:
                    self.raw_lookups.append((key, v))

            # preprocessor returns a list of filters to apply
            filters.extend(preprocessor.process_lookup(key, values))

        # at this point, need to figure out which parts of the query to "AND"
        # together and which to "OR" together ... this is naive and simply groups
        # lookups on the same column together in an OR clause.
        for filter_dict in filters:
            for field_part, value in filter_dict.items():
                if '__' in field_part:
                    column, lookup = field_part.rsplit('__', 1)
                else:
                    column = field_part

            if column not in self.exclude:
                lookups_by_column.setdefault(column, [])
                lookups_by_column[column].append(Q(**filter_dict))

        return [reduce(operator.or_, lookups) for lookups in lookups_by_column.values()]

    def get_filtered_query(self):
        nodes = self.process_request()
        if nodes:
            return self.query.filter(*nodes)

        return self.query


FIELD_TYPES = ('text', 'select', 'select_multiple', 'hidden', 'foreign_key', 'foreign_key_multiple')

class Lookup(object):
    def __init__(self, field, lookup, field_type, data=None, prefix=''):
        self.field = field
        self.lookup = lookup
        self.field_type = field_type
        self.data = data
        self.prefix = prefix

        self.name = '%s%s__%s' % (self.prefix, self.field.name, self.lookup)
        self.lookup_name = None

    def get_request(self):
        if self.lookup == 'in':
            return request.args.getlist(self.name)
        else:
            return request.args.get(self.name)

    def get_repr(self, default=''):
        if self.name in request.args:
            if self.lookup == 'in':
                pk_list = request.args.getlist(self.name)
            else:
                pk_list = [request.args.get(self.name)]
            return dict([
                (pk, unicode(self.get_by_pk(pk))) \
                    for pk in pk_list
            ])
        return default

    def get_by_pk(self, pk):
        if not isinstance(self.field, ForeignKeyField):
            raise TypeError('field %s is not a foreign key field' % self.field.name)
        model_class = self.data
        return model_class.get(**{model_class._meta.pk_name: pk})
